/*
Licensed Materials - Property of IBM
IBM WebSphere CloudBurst Appliance (9235-72X)
Copyright IBM Corporation 2009. All Rights Reserved.
US Government Users Restricted Rights - Use, duplication or disclosure
restricted by GSA ADP Schedule Contract with IBM Corp.
*/

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.ZipOutputStream;
import java.util.zip.ZipEntry;

import com.ibm.bedrock.BedrockUtils;
import com.ibm.websphere.rainmaker.logging.CleanStackTrace;
import com.ibm.ws.xsa.common.ProblemDeterminationUtils;
//import com.ibm.websphere.rest.RESTUtils;

import zero.json.Json;

def _addContentToZip(zipOut, zipEntryName, content)
{
    def entry = new ZipEntry(zipEntryName);
    zipOut.putNextEntry(entry);
    
    def bytes = content.getBytes();
    zipOut.write(bytes, 0, bytes.length);
}

def _addFileToZip(zipOut, file)
{
	if(!file.isFile()) return;
	
    def buffer = new byte[2048];
    def bytesRead = 0;
    
    file.withInputStream {
        def entry = new ZipEntry(file.name);
        zipOut.putNextEntry(entry);
        
        while ((bytesRead = it.read(buffer, 0, buffer.length)) >= 0)
            zipOut.write(buffer, 0, bytesRead);
    }
}

def _generateErrorReport()
{
    def response = executeApplianceRequest([
        name: 'soma:do-action', 
        content: [:], 
        children: [
            [
                name: 'ErrorReport', 
                content: [:], 
                children: [
                    [
                        name: 'InternalState', 
                        content: 'on'
                    ]
                ]
            ]
        ]
    ]);
    return response?.result?.equals('OK');
}

def onList()
{	
	APPLIANCE_ADMIN = 2; // define this here to get rid of the runtime dependency on r.users
	
//	if (!("xcadmin".equals(request.pdsession[]) || user.session[].permissions.contains(APPLIANCE_ADMIN)))
	//when login from from-based panel, request.pdsession is not available, but we can ensure them with valid csrf token
    if (!(com.ibm.ws.xsa.common.SecurityUtils.validateToken() || "xcadmin".equals(request.pdsession[]) || com.ibm.websphere.rest.RESTUtils.authorized()))
		return com.ibm.websphere.rest.RenderUtils.renderError(403, 'RM08008');
    
	def traceNameSuffix = "";
	 
	// Filters trace log collection to just the specified process.
	def processName = zget("/request/params/process");
	if (processName != null) {
		traceNameSuffix = "_"+processName;
	}
	
    boolean includeStats = true;
    def csv = zget("/request/params/includeStats");
    if (csv != null && csv.equalsIgnoreCase("false")) {
    	includeStats = false;
    	traceNameSuffix = traceNameSuffix+"_nostats";
    }
    
    traceNameSuffix = traceNameSuffix+"_"+_getCurrentTimeStamp();
	 
	request.headers.out.'Content-Type' = 'application/zip';
    request.headers.out.'Content-Disposition' = 'inline; filename=trace_'+zget("/request/serverName")+traceNameSuffix+'.zip';
    request.headers.out.Expires = '0';
    request.headers.out.Pragma = 'cache';
    request.headers.out.'Cache-Control' = 'private';
    
    
    def zipOut = new ZipOutputStream(request.outputStream[]);
    def diskRoot = zget("/config/rainmaker/disk/root")
    if(config.rainmaker.appName[]){ // strip off appName
    	diskRoot = diskRoot.substring(0, (diskRoot.length() -config.rainmaker.appName[].length() ))
    }

    //-----------------------------------------------------------------------------------------------------------------------------------------------
    // Leave in for legacy support.
    // look for special parms to force out dumps at some type of interval, and run jextract on cores
    // dumpProc - alone, will take a single dump of the specified process
    // numDumps - specifies the number of dumps to force, will be ignored if dumpProc is not specified
    // dumpInt - specifies the time in seconds to wait between dumps , will be ignored if dumpProc is not specified.  numDumps will default to 3.
    // deleteDumps - specifies whether to remove any dumps located after collected.
    if(zcontains("/request/params/dumpProcs")) {
    	def dumpProcName = zget("/request/params/dumpProcs");
    	def numDumps = zget("/request/params/numDumps");
    	def dumpInt = zget("/request/params/dumpInt");
    	ProblemDeterminationUtils.sendSIGQUITsignal(dumpProcName, numDumps, dumpInt);
    }
    
    // Two changes of behavior here I need to think through.
    // 1. hdump alone no longer does a heap dump on xsa.app
    // 2. proccessName set to something besides null and appliance will collect hdumps.
    if(processName != null && zcontains("/request/params/hdump")) {
        if(System.getProperty('java.vendor').startsWith('IBM')) {
            ProblemDeterminationUtils.heapDump(processName);
        }
    }
    
    def statsFileMatch = ~/.*(stats).*(log|gz)/; 
    def fileMatch = ~/.*(log|gz|txt|phd|dmp|out)/;
    if (processName != null && processName.equals("csv") || processName.equals("stats")) {
    	fileMatch = statsFileMatch;
    }
    
    try {
    	// Here we are putting the contents of many folders under diskRoot into the trace.zip file
        def diskRootDir = new File(diskRoot);
        if(diskRootDir.exists()){
        	// diskRootDirChildren is a directory such as ${diskRootDir.path}/cs-start
        	// ${diskRootDir.path}/xsServer00 on the appliance
        	def diskRootDirChildren = diskRootDir.listFiles();
        	def subdirs = ["", "error", "health", "trace", "audit", "ffdc", "0", "1"];
        	diskRootDirChildren.each(){
        		// only worry about sMash applications leveraging CB logfilehandlers for this collection
        		if(it.isDirectory() && (processName == null || processName.equals("csv") || processName.equals("stats") || Pattern.matches(processName, it.getName()))) {
        			// we are no longer making zip files out of the child directories under the diskroot
        			// ex: cs, cs-setup, xsa.app, etc.
        			// they should remain child directories at the root of trace.zip
	        		def logsDir = new File (diskRootDir, it.name + "/logs");
	        		def serverNameFile = logsDir.list(
	        			    [accept:{d, f-> f ==~ /($it.name).*/ }] as FilenameFilter
	        			  )
	        		if(serverNameFile) {
	        			subdirs.addAll(serverNameFile.toList());
	        		}
	        		
	        		try {
			            // subdirs is a directory like trace.zip/cs/logs/error
			            for (def i=0; i < subdirs.size(); i++){
			            	// logsDir is a directory like ${diskRootDir.path}/xsa.app/logs
			            	// currentLog is a directory like trace.zip/cs/logs/trace or trace.zip/xsa.app/logs/error
			                def currentLog = new File (logsDir, subdirs[i]);
			                if(currentLog.exists()) {
			                    //currentLog.eachFileMatch(~/.*(log|gz|txt|phd|dmp)/) { file ->
			                	currentLog.eachFileMatch(fileMatch) { file ->
			                    	// "file" is something like trace.zip/xsa.app/logs/trace/trace.log
			                    	String filePath = file.path;
			                    	// logsDir is a directory like ${diskRootDir.path}/xsa.app/logs

			                    	// Add file if includeStats is false and the file path is not a stats csv.
			                    	// Add file If includeStats is true continue.
			                		if ((!includeStats && !Pattern.matches(statsFileMatch.toString(), file.getName())) ||
			                				includeStats) {
				                    	filePath = filePath.substring("${diskRootDir.path}".length()+1 ); //add one to remove last separator
				                    	
				                    	// filePath is a directory like xsa.app/logs/trace/trace.log
	    			                	zipOut.putNextEntry(new ZipEntry(filePath));
	    			                	new FileInputStream(file).withStream { inStream ->
		    			                	def buffer = new byte[8192]
		    			                	def count;
		    			                	while((count = inStream.read(buffer, 0, 8192)) != -1) {
		    			                		zipOut.write(buffer,0,count)
		    			                	}
					                    }
	    			                	zipOut.closeEntry()
                                    }
                                    
                                    if (file.name.endsWith("phd")) {
                                        System.out.println("Delete file: "+file.name);
                                        file.delete();
                                    }
                                }
                            }
                        } // end traversal of subdirs
                    } catch (Throwable t) {
                        logger.SEVERE { CleanStackTrace.cleanStackTrace(t) }
                    } // end try
                } // end diskRootChildren folder traversal
            }  // end diskRootDirChildren.each()
        }

        //----------------------------------------------------------------------------------------------------------------
        // in the following sections, we add files directly to the root of trace.zip
        //----------------------------------------------------------------------------------------------------------------
        
        // [1] Get appliance.log, top, jdump, oscore
        if (processName == null || processName.equals("appliance")) {
            try {
                if(System.getProperty('java.vendor').startsWith('IBM')) {
                    
                    if(zcontains("/request/params/jdump")) {
                        com.ibm.jvm.Dump.JavaDump();
                    }
                    // always grab the javacores
                    def coreDumpDir = System.getenv("IBM_JAVACOREDIR");
                    if(coreDumpDir==null) {
                        coreDumpDir = diskRoot;
                    }
                    if(coreDumpDir != null){
                        coreDumpDir = new File(coreDumpDir);
                        coreDumpDir.eachFileMatch(~'javacore.*') { file ->
                            _addFileToZip(zipOut, file);
                        }
                    }
                }
                
                if(zcontains("/request/params/oscore")) {
                    def coreDumpDir = diskRoot;
                    if(coreDumpDir != null){
                        coreDumpDir = new File(coreDumpDir);
                        coreDumpDir.eachFileMatch(~'core(|.*)') { file ->
                            _addFileToZip(zipOut, file);
                        }
                    }
                }
                
                if(zcontains("/request/params/diskUsage") ){
                    diskRootDir = new File(diskRoot);
                    def directorySizeMap = [:];
                    _calculateDirectorySize(diskRootDir, directorySizeMap, diskRootDir);
                    
                    String directorySizeLog = diskRoot + "/logs/directorySizeDump.log";
                    File directorySizeLogFile = new File (directorySizeLog);
                    BufferedWriter bw = new BufferedWriter(new FileWriter(directorySizeLogFile));
                    zero.json.Json.encode(directorySizeMap, bw);
                    
                    _addFileToZip(zipOut, directorySizeLogFile);
                }
            } catch(Throwable e) {
                logger.SEVERE { CleanStackTrace.cleanStackTrace(e) };
            }
        }
        
	    // [2] Get global context
	    if (processName == null || processName.equals("appliance") || processName.equals("globalcontext")) {
	    	def strApplianceLogFile = "/drouter/log/appliance.log";
	    	def applianceLogFile = new File(strApplianceLogFile); 
	    	if (applianceLogFile.exists()){
	    		_addFileToZip(zipOut, applianceLogFile);
	    	}
	    	
			try {
	            // zdump('/') no longer works b/c we added a context sMash doesn't know about - _appliance.
	            // Work around is to all zdump on all the top level contexts we know about.
				//def dumpData = zdump('/')
	            def globalContexts = [
	                'appliance',
	                'config',
	                'request',
	                'storage',
	                'tmp'
	            ];
	            
	            for (globalContext in globalContexts) {
	                def dumpData = _getGlobalContext(globalContext);
	                if(dumpData!=null) {
	                    _addContentToZip(zipOut, 'global-context-'+globalContext+'-snapshot.log', dumpData);  
	                }            
	            }
			} catch(Throwable e) {
				logger.SEVERE { CleanStackTrace.cleanStackTrace(e) };
			}
	    }

		// [3] Get bedrock drouter files
		if (processName == null || processName.equals("appliance")) {
			try {			
				def packetFile = new File('/drouter/temporary/capture.pcap');
				if (packetFile.exists())
					_addFileToZip(zipOut, packetFile);
		        
				def temporaryDirectory = new File('/drouter/temporary');
				if (temporaryDirectory.exists()){
					temporaryDirectory.eachFileMatch(~/.*log/) { file ->
			            if(!file.isDirectory()){
			                _addFileToZip(zipOut, file);
					    }else{
					    	file.eachFileMatch(~/.*log/) { logFile ->
					            if(!logFile.isDirectory())
					                _addFileToZip(zipOut, logFile);
					    	}
					    }
					}
				}
				
				def drouterDir = new File('/drouter');
				def localDirectory = new File('/drouter/local');
				if (localDirectory.exists()){
					localDirectory.eachFileMatch(~/.*log/) { file ->
	            	    // "file" is something like /drouter/local/xsa.admin_firstBoot.log
	            	    def localDirFilePath = file.path;
	            	    localDirFilePath = localDirFilePath.substring("${drouterDir.path}".length() + 1);
	            	
	            	    // filePath is a directory like local/xsa.admin_firstBoot.log
	            	    zipOut.putNextEntry(new ZipEntry(localDirFilePath));
	            	
	            	    new FileInputStream(file).withStream { inStream ->
	                	    def buffer = new byte[2048];
	                	    def count;
	                	    while((count = inStream.read(buffer, 0, 2048)) != -1) {
	                		    zipOut.write(buffer);
	                	    } //end writing while loop
	                    } // end FileInputStream closure
	                	zipOut.closeEntry();
				    } // end localDirectory file-matching closure
				} //end if					
			} catch (Throwable t) {
	            logger.SEVERE { CleanStackTrace.cleanStackTrace(t) }
	        }
		}

		// [4] Get soma logs, bedrock platform collect, fio logs
		if (processName == null || processName.equals("appliance")) {
	    	try{
	    		if(onAppliance()) {
		    		def applianceStatus = executeApplianceRequest([name: 'soma:get-status',content: [ : ]]);
			    	if(applianceStatus) {
			    		applianceStatus = zero.json.Json.encode(applianceStatus);	// convert to string
			    		_addContentToZip(zipOut, 'appliance-get-status-snapshot.log', applianceStatus);
			    	}
	    		}
	    	} catch (Throwable t){
	    		logger.SEVERE { CleanStackTrace.cleanStackTrace(t) }
	    	}

            // Only do platform collect-pd if we are running on top of bedrock.
            if (BedrockUtils.isBedrock()) {
                try{
                        Runtime.getRuntime().exec("/bin/cli -e platform collect-pd").waitFor();
                        def collectPd = new File ('/tmp/userfiles/collect-pd.txt');
                        if(collectPd.exists()){
                            _addFileToZip(zipOut, collectPd);
                        }
                } catch (Throwable t){
                    logger.SEVERE { CleanStackTrace.cleanStackTrace(t) }
                }
            }
	
	    	try{
	            def initFio = new File ('/tmp/init-fio.log');
	            if(initFio.exists()){
	        	_addFileToZip(zipOut, initFio);
	            }
	    	} catch (Throwable t){
	    		logger.SEVERE { CleanStackTrace.cleanStackTrace(t) }
	    	}
		}
		
		// [5] Error reports
		if (processName == null || processName.equals("appliance")) {
	        def generatedErrorFile = new File('/drouter/temporary/error-report.txt');
	        //
	        // if no error report is present, generate a new one
	        //
	        try
	        {
	            if (!generatedErrorFile.exists())
	    		    _generateErrorReport();
	        } catch (Throwable t) {
	            logger.SEVERE { CleanStackTrace.cleanStackTrace(t) }
	        }
			
			if (generatedErrorFile.exists())
	            _addFileToZip(zipOut, generatedErrorFile);
	
	        def varLogDirectory = new File ('/var/log/');
	        if(varLogDirectory.exists()){
	        	varLogDirectory.eachFileMatch(~/(diag-log|default-trace)(|.*)/) { file ->
	            	_addFileToZip(zipOut, file);
	            }
	            varLogDirectory.eachFileMatch(~/.*\.out/) { file ->
	            	_addFileToZip(zipOut, file);
	            }
	        }
		}
    
    } catch (Throwable t) {
        return com.ibm.websphere.rest.RenderUtils.renderError(500, 'RM08009', t);
    } finally {
        zipOut.close();
    }
}

def _calculateDirectorySize(dir, directorySizeMap, diskRoot) {
	
	File[] filesInDir = dir.listFiles();
	long sizeOfFiles = 0;
	for (int i = 0; i < filesInDir.length; i++) {
		if (filesInDir[i].isDirectory()) {
			sizeOfFiles += _calculateDirectorySize(filesInDir[i], directorySizeMap, diskRoot);
		}
		else if (filesInDir[i].isFile()) {
			sizeOfFiles += filesInDir[i].length();
		}
	}
	
	// Strip off the /drouter/ramdisk2.../raid0 part of the file name and do not output directories less than 100k
	if (sizeOfFiles >= 100000) {
		directorySizeMap.put(dir.getAbsolutePath().substring(diskRoot.getAbsolutePath().length()), (int)(sizeOfFiles/1024));
	}
	
	return sizeOfFiles;
}

//
// Get the global context to print out. This method will obfuscate any data we don't
// want shared with the world, such as secret keys, passwords, etc.
//
def _getGlobalContext(globalContext) {
    def data = zdump('/' + globalContext);
    
    if ("appliance".equals(globalContext)) {
        // Update keys that don't lie within a map.
        data = _obfuscateValue(data, "/appliance/security/secretKey");
        data = _obfuscateValue(data, "/appliance/ssoTokenKey");
        data = _obfuscateValue(data, "/appliance/applianceUser");
        
        // Update keys that lie within a map.
        data = _obfuscateMapValue(data, "/appliance/security/sslProps", "authenticationSecret");
        data = _obfuscateMapValue(data, "/appliance/security/sslProps", "keyStorePassword");
        data = _obfuscateMapValue(data, "/appliance/security/sslProps", "trustStorePassword");
        data = _obfuscateMapValue(data, "/appliance/rainmaker/security/ldap", "jndiSecurityCredentials");

        // Update wildcard keys that lie within a map.
        data = _obfuscateWildcardValues(data, "/appliance/xsa/persistence/dynacaches", "accessPassword"); // "/appliance/xsa/persistence/dynacaches/*/accessPassword"
        data = _obfuscateWildcardValues(data, "/appliance/xsa/persistence/simplecaches", "accessPassword"); // "/appliance/xsa/persistence/simplecaches/*/accessPassword"
        data = _obfuscateWildcardValues(data, "/appliance/xsa/persistence/sessions", "accessPassword"); // "/appliance/xsa/persistence/sessions/*/accessPassword"
        data = _obfuscateWildcardValues(data, "/appliance/xsa/persistence/users", "passwordhash"); // "/appliance/xsa/persistence/users/*/passwordhash"
        data = _obfuscateWildcardValues(data, "/appliance/xsa/persistence/users", "salt"); // "/appliance/xsa/persistence/users/*/salt" 
        data = _obfuscateWildcardValues(data, "/appliance/tasks/tasks", "secretKey"); // "/appliance/tasks/tasks/*/taskInfo/secretKey"

    } else if ("config".equals(globalContext)) {
        // Update keys that don't lie within a map.
        data = _obfuscateValue(data, "/config/security/secretKey");
        data = _obfuscateValue(data, "/config/security/collectiveSecretKey");
        data = _obfuscateValue(data, "/config/rainmaker/admin/defaultPassword");
        
        // Update keys that lie within a map.
        data = _obfuscateMapValue(data, "/config/security/userservice/ldap", "jndiSecurityCredentials");
        data = _obfuscateMapValue(data, "/config/connection/https/defaultConfig", "keyStorePassword");
        data = _obfuscateMapValue(data, "/config/connection/https/defaultConfig", "trustStorePassword");
        data = _obfuscateMapValue(data, "/config/https/sslconfig", "keyStorePassword");
        data = _obfuscateMapValue(data, "/config/https/sslconfig", "trustStorePassword");
        data = _obfuscateMapValue(data, "/config/connection/https/managementConfig", "keyStorePassword");
        data = _obfuscateMapValue(data, "/config/connection/https/managementConfig", "trustStorePassword");
        data = _obfuscateMapValue(data, "/config/connection/smtps/defaultConfig", "trustStorePassword");
        data = _obfuscateMapValue(data, "/config/connection/https/dpHttpsConfig", "trustStorePassword");
        data = _obfuscateMapValue(data, "/config/db/rainmaker-db", "password");

    } else if ("storage".equals(globalContext)) {
        // Update keys that don't lie within a map.
        data = _obfuscateValue(data, "/storage/security/secretKey");
        
        // Update keys that lie within a map.
        data = _obfuscateMapValue(data, "/storage/security/sslProps", "authenticationSecret");
        data = _obfuscateMapValue(data, "/storage/security/sslProps", "keyStorePassword");
        data = _obfuscateMapValue(data, "/storage/security/sslProps", "trustStorePassword");

        // Update wildcard keys that lie within a map.
        data = _obfuscateWildcardValues(data, "/storage/tasks/tasks", "secretKey"); // "/storage/tasks/tasks/*/taskInfo/secretKey"
        
        // Update keys that lie within a JSON map.
        data = _obfuscateJsonMap(data, com.ibm.ws.xsa.bedrock.tasks.TasksConstants.APPLIANCE_REBOOT_TASK);
        data = _obfuscateJsonMap(data, com.ibm.ws.xsa.bedrock.tasks.TasksConstants.APPLIANCE_SHUTDOWN_TASK);
        data = _obfuscateJsonMap(data, com.ibm.ws.xsa.bedrock.tasks.TasksConstants.APPLIANCE_UPGRADE_TASK);
    }
    
    return data;
}

//
// Obfuscate a single key's value that appears in the following format:
//
//   key : value
//
def _obfuscateValue(data, key) {
    Pattern pattern = Pattern.compile('^' + key + '\\s:\\s(.*)$', Pattern.MULTILINE);
    Matcher matcher = pattern.matcher(data);

    if (matcher.find()) {
        String line = matcher.group();
        String value = matcher.group(1);
        data = data.replace(line, line.replace(value, "xxxxxx"));
    }
    
    return data;
}

//
// Obfuscate the value of a key that lies in a map:
//
//   key : {..., mapKey=value[, ...]}
//   key : {..., someKey={..., mapKey=value[, ...]}[, ...]}
//
def _obfuscateMapValue(String data, String key, String mapKey) {
    Pattern pattern = Pattern.compile('^' + key + '\\s:\\s.*(' + mapKey + '\\S+)[,\\}]+.*$', Pattern.MULTILINE);
    Matcher matcher = pattern.matcher(data);

    if (matcher.find()) {
        String line = matcher.group();
        String value = matcher.group(1);
        data = data.replace(line, line.replace(value, mapKey + "=xxxxxx"));
    }
    
    return data;
}

//
// Obfuscate the value of wildcarded key(s) that lie in a map: 
//
//   key/* : {..., mapKey=value[, ...]}
//   key/* : {..., someKey={..., mapKey=value[, ...]}[, ...]}
//
def _obfuscateWildcardValues(data, key, mapKey) {
    def list = zlistAll(key);
    if (list != null) {
        // Update all entries in map.
        for (Object listKey : list) {
            data = _obfuscateMapValue(data, listKey, mapKey);
        }
    }
    return data;
}

//
// Obfuscate values that appear in a JSON style map.
//
def _obfuscateJsonMap(data, key) {
    Pattern pattern = Pattern.compile('(?s)(?m)^' + key + '\\s:\\s(\\{.*\\}$)');
    Matcher matcher = pattern.matcher(data);

    if (matcher.find()) {
        String match = matcher.group(1);
        String edited = match.replaceAll('"secretKey":\\s".*"', '"secretKey": "xxxxxx"');
        data = data.replace(match, edited);
    }
    return data;
}

def _getCurrentTimeStamp() {
	//[2013-01-29 11:00:27:031 EST]
	SimpleDateFormat dateFormatter = new java.text.SimpleDateFormat("yyyy-MM-dd_HH_mm_ss_SSS");
	Date d = new Date(System.currentTimeMillis());
	return dateFormatter.format(d);
}
